@@ -133,10 +133,14 @@ public class Parser extends RubyObject {
      * defaults to <code>true</code>.
      *
      * <dt><code>:object_class</code>
-     * <dd>Defaults to Hash.
+     * <dd>Defaults to Hash. If another type is provided, it will be used
+     * instead of Hash to represent JSON objects. The type must respond to
+     * <code>new</code> without arguments, and return an object that respond to <code>[]=</code>.
      *
      * <dt><code>:array_class</code>
-     * <dd>Defaults to Array.
+     * <dd>Defaults to Array. If another type is provided, it will be used
+     * instead of Hash to represent JSON arrays. The type must respond to
+     * <code>new</code> without arguments, and return an object that respond to <code><<</code>.
      *
      * <dt><code>:decimal_class</code>
      * <dd>Specifies which class to use instead of the default (Float) when
@@ -144,6 +148,7 @@ public class Parser extends RubyObject {
      * in its constructor.
      * </dl>
      */
+
     @JRubyMethod(name = "new", required = 1, optional = 1, meta = true)
     public static IRubyObject newInstance(IRubyObject clazz, IRubyObject[] args, Block block) {
         Parser parser = (Parser)((RubyClass)clazz).allocate();
@@ -200,9 +205,7 @@ public class Parser extends RubyObject {
     private RubyString convertEncoding(ThreadContext context, RubyString source) {
       RubyEncoding encoding = (RubyEncoding)source.encoding(context);
       if (encoding == info.ascii8bit.get()) {
-          if (source.isFrozen()) {
-            source = (RubyString) source.dup();
-          }
+          source = (RubyString) source.dup();
           source.force_encoding(context, info.utf8.get());
       } else {
         source = (RubyString) source.encode(context, info.utf8.get());
@@ -329,7 +332,7 @@ public class Parser extends RubyObject {
 
         private RaiseException unexpectedToken(int absStart, int absEnd) {
             RubyString msg = getRuntime().newString("unexpected token at '")
-                    .cat(data, absStart, absEnd - absStart)
+                    .cat(data, absStart, Math.min(absEnd - absStart, 32))
                     .cat((byte)'\'');
             return newException(Utils.M_PARSER_ERROR, msg);
         }
@@ -339,11 +342,11 @@ public class Parser extends RubyObject {
         }
 
         
-// line 365 "Parser.rl"
+// line 368 "Parser.rl"
 
 
         
-// line 347 "Parser.java"
+// line 350 "Parser.java"
 private static byte[] init__JSON_value_actions_0()
 {
 	return new byte [] {
@@ -457,7 +460,7 @@ static final int JSON_value_error = 0;
 static final int JSON_value_en_main = 1;
 
 
-// line 471 "Parser.rl"
+// line 474 "Parser.rl"
 
 
         void parseValue(ParserResult res, int p, int pe) {
@@ -465,14 +468,14 @@ static final int JSON_value_en_main = 1;
             IRubyObject result = null;
 
             
-// line 469 "Parser.java"
+// line 472 "Parser.java"
 	{
 	cs = JSON_value_start;
 	}
 
-// line 478 "Parser.rl"
+// line 481 "Parser.rl"
             
-// line 476 "Parser.java"
+// line 479 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -498,13 +501,13 @@ case 1:
 	while ( _nacts-- > 0 ) {
 		switch ( _JSON_value_actions[_acts++] ) {
 	case 9:
-// line 456 "Parser.rl"
+// line 459 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 508 "Parser.java"
+// line 511 "Parser.java"
 		}
 	}
 
@@ -567,25 +570,25 @@ case 1:
 			switch ( _JSON_value_actions[_acts++] )
 			{
 	case 0:
-// line 373 "Parser.rl"
+// line 376 "Parser.rl"
 	{
                 result = getRuntime().getNil();
             }
 	break;
 	case 1:
-// line 376 "Parser.rl"
+// line 379 "Parser.rl"
 	{
                 result = getRuntime().getFalse();
             }
 	break;
 	case 2:
-// line 379 "Parser.rl"
+// line 382 "Parser.rl"
 	{
                 result = getRuntime().getTrue();
             }
 	break;
 	case 3:
-// line 382 "Parser.rl"
+// line 385 "Parser.rl"
 	{
                 if (parser.allowNaN) {
                     result = getConstant(CONST_NAN);
@@ -595,7 +598,7 @@ case 1:
             }
 	break;
 	case 4:
-// line 389 "Parser.rl"
+// line 392 "Parser.rl"
 	{
                 if (parser.allowNaN) {
                     result = getConstant(CONST_INFINITY);
@@ -605,7 +608,7 @@ case 1:
             }
 	break;
 	case 5:
-// line 396 "Parser.rl"
+// line 399 "Parser.rl"
 	{
                 if (pe > p + 8 &&
                     absSubSequence(p, p + 9).equals(JSON_MINUS_INFINITY)) {
@@ -634,7 +637,7 @@ case 1:
             }
 	break;
 	case 6:
-// line 422 "Parser.rl"
+// line 425 "Parser.rl"
 	{
                 parseString(res, p, pe);
                 if (res.result == null) {
@@ -647,7 +650,7 @@ case 1:
             }
 	break;
 	case 7:
-// line 432 "Parser.rl"
+// line 435 "Parser.rl"
 	{
                 currentNesting++;
                 parseArray(res, p, pe);
@@ -662,7 +665,7 @@ case 1:
             }
 	break;
 	case 8:
-// line 444 "Parser.rl"
+// line 447 "Parser.rl"
 	{
                 currentNesting++;
                 parseObject(res, p, pe);
@@ -676,7 +679,7 @@ case 1:
                 }
             }
 	break;
-// line 680 "Parser.java"
+// line 683 "Parser.java"
 			}
 		}
 	}
@@ -696,7 +699,7 @@ case 5:
 	break; }
 	}
 
-// line 479 "Parser.rl"
+// line 482 "Parser.rl"
 
             if (cs >= JSON_value_first_final && result != null) {
                 if (parser.freeze) {
@@ -709,7 +712,7 @@ case 5:
         }
 
         
-// line 713 "Parser.java"
+// line 716 "Parser.java"
 private static byte[] init__JSON_integer_actions_0()
 {
 	return new byte [] {
@@ -808,7 +811,7 @@ static final int JSON_integer_error = 0;
 static final int JSON_integer_en_main = 1;
 
 
-// line 501 "Parser.rl"
+// line 504 "Parser.rl"
 
 
         void parseInteger(ParserResult res, int p, int pe) {
@@ -826,15 +829,15 @@ static final int JSON_integer_en_main = 1;
             int cs = EVIL;
 
             
-// line 830 "Parser.java"
+// line 833 "Parser.java"
 	{
 	cs = JSON_integer_start;
 	}
 
-// line 518 "Parser.rl"
+// line 521 "Parser.rl"
             int memo = p;
             
-// line 838 "Parser.java"
+// line 841 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -915,13 +918,13 @@ case 1:
 			switch ( _JSON_integer_actions[_acts++] )
 			{
 	case 0:
-// line 495 "Parser.rl"
+// line 498 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 925 "Parser.java"
+// line 928 "Parser.java"
 			}
 		}
 	}
@@ -941,7 +944,7 @@ case 5:
 	break; }
 	}
 
-// line 520 "Parser.rl"
+// line 523 "Parser.rl"
 
             if (cs < JSON_integer_first_final) {
                 return -1;
@@ -961,7 +964,7 @@ case 5:
         }
 
         
-// line 965 "Parser.java"
+// line 968 "Parser.java"
 private static byte[] init__JSON_float_actions_0()
 {
 	return new byte [] {
@@ -1063,7 +1066,7 @@ static final int JSON_float_error = 0;
 static final int JSON_float_en_main = 1;
 
 
-// line 553 "Parser.rl"
+// line 556 "Parser.rl"
 
 
         void parseFloat(ParserResult res, int p, int pe) {
@@ -1082,15 +1085,15 @@ static final int JSON_float_en_main = 1;
             int cs = EVIL;
 
             
-// line 1086 "Parser.java"
+// line 1089 "Parser.java"
 	{
 	cs = JSON_float_start;
 	}
 
-// line 571 "Parser.rl"
+// line 574 "Parser.rl"
             int memo = p;
             
-// line 1094 "Parser.java"
+// line 1097 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -1171,13 +1174,13 @@ case 1:
 			switch ( _JSON_float_actions[_acts++] )
 			{
 	case 0:
-// line 544 "Parser.rl"
+// line 547 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 1181 "Parser.java"
+// line 1184 "Parser.java"
 			}
 		}
 	}
@@ -1197,7 +1200,7 @@ case 5:
 	break; }
 	}
 
-// line 573 "Parser.rl"
+// line 576 "Parser.rl"
 
             if (cs < JSON_float_first_final) {
                 return -1;
@@ -1207,7 +1210,7 @@ case 5:
         }
 
         
-// line 1211 "Parser.java"
+// line 1214 "Parser.java"
 private static byte[] init__JSON_string_actions_0()
 {
 	return new byte [] {
@@ -1309,7 +1312,7 @@ static final int JSON_string_error = 0;
 static final int JSON_string_en_main = 1;
 
 
-// line 612 "Parser.rl"
+// line 615 "Parser.rl"
 
 
         void parseString(ParserResult res, int p, int pe) {
@@ -1317,15 +1320,15 @@ static final int JSON_string_en_main = 1;
             IRubyObject result = null;
 
             
-// line 1321 "Parser.java"
+// line 1324 "Parser.java"
 	{
 	cs = JSON_string_start;
 	}
 
-// line 619 "Parser.rl"
+// line 622 "Parser.rl"
             int memo = p;
             
-// line 1329 "Parser.java"
+// line 1332 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -1406,7 +1409,7 @@ case 1:
 			switch ( _JSON_string_actions[_acts++] )
 			{
 	case 0:
-// line 587 "Parser.rl"
+// line 590 "Parser.rl"
 	{
                 int offset = byteList.begin();
                 ByteList decoded = decoder.decode(byteList, memo + 1 - offset,
@@ -1421,13 +1424,13 @@ case 1:
             }
 	break;
 	case 1:
-// line 600 "Parser.rl"
+// line 603 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 1431 "Parser.java"
+// line 1434 "Parser.java"
 			}
 		}
 	}
@@ -1447,7 +1450,7 @@ case 5:
 	break; }
 	}
 
-// line 621 "Parser.rl"
+// line 624 "Parser.rl"
 
             if (parser.createAdditions) {
                 RubyHash matchString = parser.match_string;
@@ -1492,7 +1495,7 @@ case 5:
         }
 
         
-// line 1496 "Parser.java"
+// line 1499 "Parser.java"
 private static byte[] init__JSON_array_actions_0()
 {
 	return new byte [] {
@@ -1605,7 +1608,7 @@ static final int JSON_array_error = 0;
 static final int JSON_array_en_main = 1;
 
 
-// line 701 "Parser.rl"
+// line 704 "Parser.rl"
 
 
         void parseArray(ParserResult res, int p, int pe) {
@@ -1625,14 +1628,14 @@ static final int JSON_array_en_main = 1;
             }
 
             
-// line 1629 "Parser.java"
+// line 1632 "Parser.java"
 	{
 	cs = JSON_array_start;
 	}
 
-// line 720 "Parser.rl"
+// line 723 "Parser.rl"
             
-// line 1636 "Parser.java"
+// line 1639 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -1713,7 +1716,7 @@ case 1:
 			switch ( _JSON_array_actions[_acts++] )
 			{
 	case 0:
-// line 670 "Parser.rl"
+// line 673 "Parser.rl"
 	{
                 parseValue(res, p, pe);
                 if (res.result == null) {
@@ -1730,13 +1733,13 @@ case 1:
             }
 	break;
 	case 1:
-// line 685 "Parser.rl"
+// line 688 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 1740 "Parser.java"
+// line 1743 "Parser.java"
 			}
 		}
 	}
@@ -1756,7 +1759,7 @@ case 5:
 	break; }
 	}
 
-// line 721 "Parser.rl"
+// line 724 "Parser.rl"
 
             if (cs >= JSON_array_first_final) {
                 res.update(result, p + 1);
@@ -1766,7 +1769,7 @@ case 5:
         }
 
         
-// line 1770 "Parser.java"
+// line 1773 "Parser.java"
 private static byte[] init__JSON_object_actions_0()
 {
 	return new byte [] {
@@ -1889,7 +1892,7 @@ static final int JSON_object_error = 0;
 static final int JSON_object_en_main = 1;
 
 
-// line 778 "Parser.rl"
+// line 781 "Parser.rl"
 
 
         void parseObject(ParserResult res, int p, int pe) {
@@ -1914,14 +1917,14 @@ static final int JSON_object_en_main = 1;
             }
 
             
-// line 1918 "Parser.java"
+// line 1921 "Parser.java"
 	{
 	cs = JSON_object_start;
 	}
 
-// line 802 "Parser.rl"
+// line 805 "Parser.rl"
             
-// line 1925 "Parser.java"
+// line 1928 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -2002,7 +2005,7 @@ case 1:
 			switch ( _JSON_object_actions[_acts++] )
 			{
 	case 0:
-// line 735 "Parser.rl"
+// line 738 "Parser.rl"
 	{
                 parseValue(res, p, pe);
                 if (res.result == null) {
@@ -2019,7 +2022,7 @@ case 1:
             }
 	break;
 	case 1:
-// line 750 "Parser.rl"
+// line 753 "Parser.rl"
 	{
                 parseString(res, p, pe);
                 if (res.result == null) {
@@ -2037,13 +2040,13 @@ case 1:
             }
 	break;
 	case 2:
-// line 766 "Parser.rl"
+// line 769 "Parser.rl"
 	{
                 p--;
                 { p += 1; _goto_targ = 5; if (true)  continue _goto;}
             }
 	break;
-// line 2047 "Parser.java"
+// line 2050 "Parser.java"
 			}
 		}
 	}
@@ -2063,7 +2066,7 @@ case 5:
 	break; }
 	}
 
-// line 803 "Parser.rl"
+// line 806 "Parser.rl"
 
             if (cs < JSON_object_first_final) {
                 res.update(null, p + 1);
@@ -2096,7 +2099,7 @@ case 5:
         }
 
         
-// line 2100 "Parser.java"
+// line 2103 "Parser.java"
 private static byte[] init__JSON_actions_0()
 {
 	return new byte [] {
@@ -2199,7 +2202,7 @@ static final int JSON_error = 0;
 static final int JSON_en_main = 1;
 
 
-// line 854 "Parser.rl"
+// line 857 "Parser.rl"
 
 
         public IRubyObject parseImplemetation() {
@@ -2209,16 +2212,16 @@ static final int JSON_en_main = 1;
             ParserResult res = new ParserResult();
 
             
-// line 2213 "Parser.java"
+// line 2216 "Parser.java"
 	{
 	cs = JSON_start;
 	}
 
-// line 863 "Parser.rl"
+// line 866 "Parser.rl"
             p = byteList.begin();
             pe = p + byteList.length();
             
-// line 2222 "Parser.java"
+// line 2225 "Parser.java"
 	{
 	int _klen;
 	int _trans = 0;
@@ -2299,7 +2302,7 @@ case 1:
 			switch ( _JSON_actions[_acts++] )
 			{
 	case 0:
-// line 840 "Parser.rl"
+// line 843 "Parser.rl"
 	{
                 parseValue(res, p, pe);
                 if (res.result == null) {
@@ -2311,7 +2314,7 @@ case 1:
                 }
             }
 	break;
-// line 2315 "Parser.java"
+// line 2318 "Parser.java"
 			}
 		}
 	}
@@ -2331,7 +2334,7 @@ case 5:
 	break; }
 	}
 
-// line 866 "Parser.rl"
+// line 869 "Parser.rl"
 
             if (cs >= JSON_first_final && p == pe) {
                 return result;
